# regression test for testcase_dep2str of deeply nested dependencies
# this used to segfault or return wrong roundtrip results
genid dep a | b | c | d | e | f | g | h | i | j | k | l | m | n | o | p | q | r | s | t | u | v | w | x | y |z
result genid <inline>
#>genid  1: genid lit a
#>genid  2: genid lit b
#>genid  3: genid lit c
#>genid  4: genid lit d
#>genid  5: genid lit e
#>genid  6: genid lit f
#>genid  7: genid lit g
#>genid  8: genid lit h
#>genid  9: genid lit i
#>genid 10: genid lit j
#>genid 11: genid lit k
#>genid 12: genid lit l
#>genid 13: genid lit m
#>genid 14: genid lit n
#>genid 15: genid lit o
#>genid 16: genid lit p
#>genid 17: genid lit q
#>genid 18: genid lit r
#>genid 19: genid lit s
#>genid 20: genid lit t
#>genid 21: genid lit u
#>genid 22: genid lit v
#>genid 23: genid lit w
#>genid 24: genid lit x
#>genid 25: genid lit y
#>genid 26: genid lit z
#>genid 27: genid op |
#>genid 28: genid op |
#>genid 29: genid op |
#>genid 30: genid op |
#>genid 31: genid op |
#>genid 32: genid op |
#>genid 33: genid op |
#>genid 34: genid op |
#>genid 35: genid op |
#>genid 36: genid op |
#>genid 37: genid op |
#>genid 38: genid op |
#>genid 39: genid op |
#>genid 40: genid op |
#>genid 41: genid op |
#>genid 42: genid op |
#>genid 43: genid op |
#>genid 44: genid op |
#>genid 45: genid op |
#>genid 46: genid op |
#>genid 47: genid op |
#>genid 48: genid op |
#>genid 49: genid op |
#>genid 50: genid op |
#>genid 51: genid op |
#>genid dep a | b | c | d | e | f | g | h | i | j | k | l | m | n | o | p | q | r | s | t | u | v | w | x | y | z
